@@ -25,22 +25,18 @@
 
 from __future__ import print_function
 import sys
-import decimal
 import socket
-import struct
-import hashlib
-import binascii
+import decimal
 import datetime
-import re
 
-PY2 = sys.version_info[0] == 2
-
-VERSION = (0, 2, 2)
+VERSION = (0, 2, 3)
 __version__ = '%s.%s.%s' % VERSION
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'format'
 
+PY2 = sys.version_info[0] == 2
+
 DEBUG = False
 
 def DEBUG_OUTPUT(*argv):
@@ -50,40 +46,12 @@ def DEBUG_OUTPUT(*argv):
         print(s, end=' ', file=sys.stdout)
     print(file=sys.stdout)
 
-Date = datetime.date
-Time = datetime.time
-TimeDelta = datetime.timedelta
-Timestamp = datetime.datetime
-def DateFromTicks(ticks):
-    return apply(Date,time.localtime(ticks)[:3])
-def TimeFromTicks(ticks):
-    return apply(Time,time.localtime(ticks)[3:6])
-def TimestampFromTicks(ticks):
-    return apply(Timestamp,time.localtime(ticks)[:6])
-def Binary(b):
-    return bytes(b)
+def HEX(data):
+    import binascii
+    return binascii.b2a_hex(data)
 
-class DBAPITypeObject:
-    def __init__(self,*values):
-        self.values = values
-    def __cmp__(self,other):
-        if other in self.values:
-            return 0
-        if other < self.values:
-            return 1
-        else:
-            return -1
-STRING = DBAPITypeObject(str)
-BINARY = DBAPITypeObject(bytes)
-NUMBER = DBAPITypeObject(int, decimal.Decimal)
-DATETIME = DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
-DATE = DBAPITypeObject(datetime.date)
-TIME = DBAPITypeObject(datetime.time)
-ROWID = DBAPITypeObject()
-
-# Format Codes
-FC_BINARY = 1
-FC_TEXT = 0
+#-----------------------------------------------------------------------------
+# http://www.postgresql.org/docs/9.3/static/protocol.html
 
 # http://www.postgresql.org/docs/9.3/static/protocol-message-formats.html
 PG_B_AUTHENTICATION = b'R'
@@ -207,32 +175,17 @@ PG_TYPE_ANYENUM = 3500
 PG_TYPE_FDW_HANDLER = 3115
 PG_TYPE_ANYRANGE = 3831
 
-def _bytes_to_bint(b, u=False):     # Read as big endian
-    if u:
-        fmtmap = {1: 'B', 2: '>H', 4: '>L', 8: '>Q'}
-    else:
-        fmtmap = {1: 'b', 2: '>h', 4: '>l', 8: '>q'}
-    fmt = fmtmap.get(len(b))
-    if fmt is None:
-        raise InternalError
-    return struct.unpack(fmt, b)[0]
-
-def _bint_to_bytes(val, nbytes):    # Convert int value to big endian bytes.
-    v = abs(val)
-    b = []
-    for n in range(nbytes):
-        b.append((v >> (8*(nbytes - n - 1)) & 0xff))
-    if val < 0:
-        for i in range(nbytes):
-            b[i] = ~b[i] + 256
-        b[-1] += 1
-        for i in range(nbytes):
-            if b[nbytes -i -1] == 256:
-                b[nbytes -i -1] = 0
-                b[nbytes -i -2] += 1
-    return b''.join([chr(c) for c in b]) if PY2 else bytes(b)
-
 def _decode_column(data, oid, encoding):
+    class UTC(datetime.tzinfo):
+        def utcoffset(self, dt):
+            return datetime.timedelta(0)
+
+        def dst(self, dt):
+            return datetime.timedelta(0)
+
+        def tzname(self, dt):
+            return "UTC"
+
     if data is None:
         return data
     data = data.decode(encoding)
@@ -296,14 +249,90 @@ def _decode_column(data, oid, encoding):
     DEBUG_OUTPUT('_decode_column()', data, oid)
     return data
 
+# ----------------------------------------------------------------------------
+
+def _bytes_to_bint(b):     # Read as big endian
+    r = b[0]
+    if PY2:
+        r = ord(r)
+    for i in b[1:]:
+        if PY2:
+            i = ord(i)
+        r = r * 256 + i
+    return r
+
+def _bint_to_bytes(val, nbytes):    # Convert int value to big endian bytes.
+    v = abs(val)
+    b = []
+    for n in range(nbytes):
+        b.append((v >> (8*(nbytes - n - 1)) & 0xff))
+    if val < 0:
+        for i in range(nbytes):
+            b[i] = ~b[i] + 256
+        b[-1] += 1
+        for i in range(nbytes):
+            if b[nbytes -i -1] == 256:
+                b[nbytes -i -1] = 0
+                b[nbytes -i -2] += 1
+    return b''.join([chr(c) for c in b]) if PY2 else bytes(b)
+
+def escape_parameter(v):
+    t = type(v)
+    if v is None:
+        return 'NULL'
+    elif (PY2 and t == unicode) or (not PY2 and t == str):  # string
+        return u"'" + v.replace(u"'", u"''") + u"'"
+    elif PY2 and t == str:    # PY2 str
+        v = ''.join(['\\%03o' % (ord(c), ) if ord(c) < 32 or ord(c) > 127 else c for c in v])
+        return "'" + v.replace("'", "''") + "'"
+    elif t == bytearray or t == bytes:        # binary
+        return "'" + ''.join(['\\%03o' % (c, ) for c in v]) + "'::bytea"
+    elif t == bool:
+        return u"'t'" if v else u"'f'"
+    elif t == int or t == float or t == decimal.Decimal or (PY2 and t == long):
+        return str(v)
+    else:
+        return "'" + str(v) + "'"
+
+
+Date = datetime.date
+Time = datetime.time
+TimeDelta = datetime.timedelta
+Timestamp = datetime.datetime
+def DateFromTicks(ticks):
+    return apply(Date,time.localtime(ticks)[:3])
+def TimeFromTicks(ticks):
+    return apply(Time,time.localtime(ticks)[3:6])
+def TimestampFromTicks(ticks):
+    return apply(Timestamp,time.localtime(ticks)[:6])
+def Binary(b):
+    return bytes(b)
+
+class DBAPITypeObject:
+    def __init__(self,*values):
+        self.values = values
+    def __cmp__(self,other):
+        if other in self.values:
+            return 0
+        if other < self.values:
+            return 1
+        else:
+            return -1
+STRING = DBAPITypeObject(str)
+BINARY = DBAPITypeObject(bytes)
+NUMBER = DBAPITypeObject(int, decimal.Decimal)
+DATETIME = DBAPITypeObject(datetime.datetime, datetime.date, datetime.time)
+DATE = DBAPITypeObject(datetime.date)
+TIME = DBAPITypeObject(datetime.time)
+ROWID = DBAPITypeObject()
+
 class Error(Exception):
-    def __init__(self, errcode, message):
-        self._errcode = errcode
-        self._message = message
+    def __init__(self, message):
+        self.message = message
     def __str__(self):
-        return self._errcode + u":" + self._message
+        return self.message
     def __repr__(self):
-        return self._errcode + u":" + self._message
+        return self.message
 
 class InterfaceError(Error):
     pass
@@ -334,35 +363,6 @@ class NotSupportedError(DatabaseError):
     def __init__(self):
         DatabaseError.__init__(self, 'NotSupportedError')
 
-class UTC(datetime.tzinfo):
-    def utcoffset(self, dt):
-        return datetime.timedelta(0)
-
-    def dst(self, dt):
-        return datetime.timedelta(0)
-
-    def tzname(self, dt):
-        return "UTC"
-
-#------------------------------------------------------------------------------
-def escape_parameter(v):
-    t = type(v)
-    if v is None:
-        return 'NULL'
-    elif (PY2 and t == unicode) or (not PY2 and t == str):  # string
-        return u"'" + v.replace(u"'", u"''") + u"'"
-    elif PY2 and t == str:    # PY2 str
-        v = ''.join(['\\%03o' % (ord(c), ) if ord(c) < 32 or ord(c) > 127 else c for c in v])
-        return "'" + v.replace("'", "''") + "'"
-    elif t == bytearray or t == bytes:        # binary
-        return "'" + ''.join(['\\%03o' % (c, ) for c in v]) + "'::bytea"
-    elif t == bool:
-        return u"'t'" if v else u"'f'"
-    elif t == int or t == float or t == long:
-        return str(v)
-    else:
-        return "'" + str(v) + "'"
-
 class Cursor(object):
     def __init__(self, connection):
         self.connection = connection
@@ -423,7 +423,6 @@ class Cursor(object):
     def next(self):
         return self.__next__()
 
-
 class Connection(object):
     def __init__(self, user, password, database, host, port, timeout, use_ssl):
         DEBUG_OUTPUT("Connection::__init__()")
@@ -439,12 +438,10 @@ class Connection(object):
         self.in_transaction = False
         self._open()
 
-
     def _send_message(self, code, data):
         self._write(
             b''.join([code, _bint_to_bytes(len(data) + 4, 4), data, PG_F_FLUSH, b'\x00\x00\x00\x04'])
         )
-        self._flush()
 
     def _process_messages(self, obj=None):
         err = None
@@ -462,6 +459,7 @@ class Connection(object):
                 if auth_method == 0:      # trust
                     pass
                 elif auth_method == 5:    # md5
+                    import hashlib
                     salt = data[4:]
                     hash1 = hashlib.md5(self.password.encode('ascii') + self.user.encode("ascii")).hexdigest().encode("ascii")
                     hash2 = hashlib.md5(hash1+salt).hexdigest().encode("ascii")
@@ -476,7 +474,7 @@ class Connection(object):
                 if k == 'server_encoding':
                     self.encoding = v
             elif code == PG_B_BACKEND_KEY_DATA:
-                DEBUG_OUTPUT("BACKEND_KEY_DATA:", binascii.b2a_hex(data))
+                DEBUG_OUTPUT("BACKEND_KEY_DATA:", HEX(data))
             elif code == PG_B_COMMAND_COMPLETE:
                 command = data[:-1].decode('ascii')
                 DEBUG_OUTPUT("COMMAND_COMPLETE:", command)
@@ -490,7 +488,7 @@ class Connection(object):
                             obj._current_row = -1
                             break
             elif code == PG_B_ROW_DESCRIPTION:
-                DEBUG_OUTPUT("ROW_DESCRIPTION:", binascii.b2a_hex(data))
+                DEBUG_OUTPUT("ROW_DESCRIPTION:", HEX(data))
                 count = _bytes_to_bint(data[0:2])
                 obj.description = [None] * count
                 n = 2
@@ -516,15 +514,16 @@ class Connection(object):
                     idx += 1
                 DEBUG_OUTPUT('\t\t', obj.description)
             elif code == PG_B_DATA_ROW:
-                DEBUG_OUTPUT("DATA_ROW:", binascii.b2a_hex(data))
+                DEBUG_OUTPUT("DATA_ROW:", HEX(data))
                 n = 2
                 row = []
                 while n < len(data):
-                    ln = _bytes_to_bint(data[n:n+4])
-                    n += 4
-                    if ln == -1:
+                    if data[n:n+4] == b'\xff\xff\xff\xff':
                         row.append(None)
+                        n += 4
                     else:
+                        ln = _bytes_to_bint(data[n:n+4])
+                        n += 4
                         row.append(data[n:n+ln])
                         n += ln
                 for i in range(len(row)):
@@ -533,21 +532,21 @@ class Connection(object):
                 obj._rows.append(tuple(row))
                 DEBUG_OUTPUT("\t\t", row)
             elif code == PG_B_NOTICE_RESPONSE:
-                DEBUG_OUTPUT("NOTICE_RESPONSE:", binascii.b2a_hex(data))
+                DEBUG_OUTPUT("NOTICE_RESPONSE:", HEX(data))
                 for s in data.split(b'\x00'):
                     DEBUG_OUTPUT("\t\t", s.decode(self.encoding))
             elif code == PG_B_ERROR_RESPONSE:
-                DEBUG_OUTPUT("ERROR_RESPONSE:", binascii.b2a_hex(data))
+                DEBUG_OUTPUT("ERROR_RESPONSE:", HEX(data))
                 err = data.split(b'\x00')
                 for b in err:
                     DEBUG_OUTPUT("\t\t", b.decode(self.encoding))
                 # http://www.postgresql.org/docs/9.3/static/errcodes-appendix.html
                 errcode = err[1][1:].decode(self.encoding)
-                message = err[2][1:].decode(self.encoding)
+                message = errcode + u':' + err[2][1:].decode(self.encoding)
                 if errcode[:2] == u'23':
-                    err = IntegrityError(errcode, message)
+                    err = IntegrityError( message)
                 else:
-                    err = ProgrammingError(errcode, message)
+                    err = DatabaseError(message)
             elif code == PG_B_COPY_OUT_RESPONSE:
                 is_binary = data[0] == '\x01'
                 num_columns = _bytes_to_bint(data[1:3])
@@ -567,11 +566,9 @@ class Connection(object):
                         break
                     self._write(PG_COPY_DATA + _bint_to_bytes(len(buf) + 4, 4))
                     self._write(buf)
-                    self._flush()
                 self._write(PG_COPY_DONE + b'\x00\x00\x00\x04' + PG_F_SYNC + b'\x00\x00\x00\x04')
-                self._flush()
             else:
-                DEBUG_OUTPUT("SKIP:", code, ln, binascii.b2a_hex(data))
+                DEBUG_OUTPUT("SKIP:", code, ln, HEX(data))
         if err:
             raise err
         return
@@ -583,13 +580,19 @@ class Connection(object):
         self.close()
 
     def _read(self, ln):
-        return self.sock.recv(ln)
+        if not self.sock:
+            raise OperationalError(u"08003:Lost connection")
+        r = b''
+        while len(r) < ln:
+            r += self.sock.recv(ln-len(r))
+        return r
 
     def _write(self, b):
-        return self.sock.send(b)
-
-    def _flush(self):
-        pass
+        if not self.sock:
+            raise OperationalError(u"08003:Lost connection")
+        n = 0
+        while (n < len(b)):
+            n += self.sock.send(b[n:])
 
     def _open(self):
         self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
