@@ -23,10 +23,10 @@
 #OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 #SOFTWARE.
 ##############################################################################
-import minipg
 import unittest
-import minipg
 import io
+import decimal
+import minipg
 
 class TestMiniPG(unittest.TestCase):
     host='localhost'
@@ -80,12 +80,14 @@ class TestMiniPG(unittest.TestCase):
         self.assertEqual(cur.fetchone()[0], 1)
         cur.execute("select count(*) from test_basic where b1=%s", (False,))
         self.assertEqual(cur.fetchone()[0], 2)
+        cur.execute("select i2 from test_basic where dec=%s", (decimal.Decimal("1.1"),))
+        self.assertEqual(cur.fetchone()[0], 1)
 
         cur.execute("select to_json(test_basic) from test_basic")
         self.assertEqual(len(cur.fetchall()), 3)
         try:
             cur.execute("E")
-        except minipg.ProgrammingError as e:
+        except minipg.DatabaseError as e:
             self.assertEqual(str(e), u'42601:syntax error at or near "E"')
 
     def test_trans(self):
